#ifndef FEATURE_MATCH_POLYLINE_2D_H_
#define FEATURE_MATCH_POLYLINE_2D_H_

// Wolf includes
#include "core/common/wolf.h"
#include "core/feature/feature_match.h"

//wolf nampseace
namespace wolf {
    
//forward declaration to typedef class pointers
WOLF_PTR_TYPEDEFS(FeaturePolyline2D);
WOLF_STRUCT_PTR_TYPEDEFS(FeatureMatchPolyline2D);

/** \brief Match between a feature_polyline_2D and a feature_polyline_2D
 *
 * Match between a feature_polyline_2D and a feature_polyline_2D (feature-feature correspondence)
 *
 */
struct FeatureMatchPolyline2D : public FeatureMatch
{
    int feature_last_from_id_;
    int feature_last_to_id_;
    int feature_incoming_from_id_;
    int feature_incoming_to_id_;
    Eigen::Matrix3s T_incoming_last_;
    FeaturePolyline2DPtr pl_incoming_;
    FeaturePolyline2DPtr pl_last_;

    bool check(bool check_partial_match = true) const;

    /* compute square distances
     *
     * Returns an array containing the squared distances of each point of the match
     */
    Eigen::Array<Scalar,Eigen::Dynamic,1> computeSqDistArray() const;

    void print() const;

    bool fromDefined() const;

    bool toDefined() const;

    unsigned int getNPoints() const;

    unsigned int getNDefinedPoints() const;
};

}//end namespace

#endif

