#ifndef LANDMARK_MATCH_POLYLINE_2D_H_
#define LANDMARK_MATCH_POLYLINE_2D_H_

// Wolf includes
#include "core/common/wolf.h"
#include "core/landmark/landmark_match.h"

//wolf nampseace
namespace wolf {

WOLF_PTR_TYPEDEFS(FeaturePolyline2D);
WOLF_PTR_TYPEDEFS(LandmarkPolyline2D);

// Map of Feature - Landmark matches
WOLF_STRUCT_PTR_TYPEDEFS(LandmarkMatchPolyline2D);
typedef std::map<FeatureBasePtr, LandmarkMatchPolyline2DPtr> LandmarkMatchPolyline2DMap;

/** \brief Match between a feature_polyline_2D and a landmark_polyline_2D
 *
 * Match between a feature_polyline_2D and a landmark_polyline_2D
 *
 **/
struct LandmarkMatchPolyline2D : public LandmarkMatch
{
    int landmark_from_id_;
    int feature_from_id_;
    int landmark_to_id_;
    int feature_to_id_;
    int landmark_version_;
    Eigen::Matrix3s T_feature_landmark_;
    FeaturePolyline2DPtr pl_feature_;
    LandmarkPolyline2DPtr pl_landmark_;

    bool check(bool check_partial_match = true) const;

    /* is a partial match
     *
     * Returns if there are not matched points for both landmark and feature in any or both extremes
     */
    bool isPartialFront() const;
    bool isPartialBack() const;

    /* is a deprecated match
     *
     * Returns if the landmark version is newer than when matched or if landmark was merged in another landmark
     */
    bool isDeprecated() const;

    /* compute square distances
     *
     * Returns an array containing the squared distances of each point of the match
     */
    Eigen::Array<Scalar,Eigen::Dynamic,1> computeSqDistArray() const;

    void print() const;

    bool fromDefined() const;

    bool toDefined() const;

    unsigned int getNPoints() const;

    unsigned int getNDefinedPoints() const;
};

}//end namespace

#endif
