/*
 * polyline_2D_utils.h
 *
 *  Created on: Mar 14, 2019
 *      Author: jvallve
 */

#ifndef POLYLINE_2D_UTILS_H_
#define POLYLINE_2D_UTILS_H_

#include "core/common/wolf.h"
#include "laser/landmark/landmark_match_polyline_2D.h"
#include "laser/landmark/landmark_polyline_2D.h"
#include "laser/feature/feature_match_polyline_2D.h"
#include "laser/feature/feature_polyline_2D.h"
#include "core/math/rotations.h"

namespace wolf {

WOLF_STRUCT_PTR_TYPEDEFS(MatchPolyline2D);
typedef std::map<Scalar,MatchPolyline2DPtr>    MatchPolyline2DMap;
typedef std::list<MatchPolyline2DPtr>          MatchPolyline2DList;

/** \brief Match between a two polylines (not specialized to landmark nor feature)
 *
 * Match between a two polylines (not specialized to landmark nor feature)
 *
 **/
struct MatchPolyline2D
{
    int from_A_id_;
    int to_A_id_;
    int from_B_id_;
    int to_B_id_;
    Scalar normalized_score_;
    Scalar overlap_dist_;
    Eigen::Matrix3s T_A_B_;

    void print() const
    {
        std::cout << std::endl << "POLYLINE2D MATCH" << std::endl;
        std::cout << "A: " << std::endl;
        std::cout << "\tfrom_A_id_ " << from_A_id_ << std::endl;
        std::cout << "\tto_A_id_ " << to_A_id_ << std::endl;

        std::cout << "B " << std::endl;
        std::cout << "\tfrom_B_id_ " << from_B_id_ << std::endl;
        std::cout << "\tto_B_id_ " << to_B_id_ << std::endl;

        std::cout << "MATCH " << std::endl;
        std::cout << "\tnormalized_score_ " << normalized_score_ << std::endl;
        std::cout << "\toverlap_dist_ " << overlap_dist_ << std::endl;
        std::cout << "\tT_A_B_ " << std::endl << T_A_B_ << std::endl;
    }
};

Eigen::Matrix3s pose2T2D(const Eigen::Vector2s& t, const Scalar& theta);
Eigen::Matrix3s pose2T2D(const Eigen::Vector3s& pose);
Eigen::Vector3s T2pose2D(const Eigen::Matrix3s& T);

Eigen::Vector3s computeRigidTrans(const Eigen::MatrixXs& _points_incoming, const Eigen::MatrixXs& _points_last, const bool& first_defined = true, const bool& last_defined=true);

/** \brief Computes the squared distance from a point to a segment defined by two points
 **/
Scalar sqDistPoint2Segment(const Eigen::Vector2s& p, const Eigen::Vector2s& a, const Eigen::Vector2s& b);

Scalar sqDistPointToLine(const Eigen::Vector3s& _A, const Eigen::Vector3s& _A_aux,
                         const Eigen::Vector3s& _B, bool _A_defined, bool _B_defined, bool strict=false);

MatchPolyline2DPtr computeBestSqDist(const Eigen::MatrixXs& _points_A, bool _first_defined_A, bool _last_defined_A, bool _closed_A,
                                     const Eigen::MatrixXs& _points_B, bool _first_defined_B, bool _last_defined_B, bool _closed_B,
                                     const Scalar& max_sq_error,
                                     const Scalar& min_overlapping_dist,
                                     int min_N_overlapped,
                                     int min_N_defined_overlapped,
                                     bool both_exceeding_A_allowed_ = true,
                                     bool both_exceeding_B_allowed_ = true,
									 bool print = false);

Eigen::Array<Scalar,Eigen::Dynamic,1> computeSqDist(const Eigen::MatrixXs& _points_A, const int& from_A, const int& to_A, bool from_A_defined, bool to_A_defined,
                                                    const Eigen::MatrixXs& _points_B, const int& from_B, const int& to_B, bool from_B_defined, bool to_B_defined,
                                                    const int& last_A, const int& N_overlapped );

MatchPolyline2DList computeBestRigidTrans(const Eigen::MatrixXs& _points_A, bool _first_defined_A, bool _last_defined_A, bool _closed_A,
                                          const Eigen::MatrixXs& _points_B, bool _first_defined_B, bool _last_defined_B, bool _closed_B,
                                          const Scalar& max_sq_error );

bool isProjectedOverSegment(const Eigen::Vector3s& _A, const Eigen::Vector3s& _A_aux, const Eigen::Vector3s& _B);

Eigen::VectorXs computeSquaredMahalanobisDistances(const Eigen::Vector2s& _feature,
                                                   const Eigen::Matrix2s& _feature_cov,
                                                   const Eigen::Vector2s& _expected_feature,
                                                   const Eigen::Matrix2s& _expected_feature_cov,
                                                   const Eigen::MatrixXs& _mu);

} // namespace wolf

#endif /* POLYLINE_2D_UTILS_H_ */
