/*
 * processor_tracker_feature_polyline_2D.h
 *
 *  Created on: Mar 11, 2019
 *      Author: jvallve
 */

#ifndef PROCESSOR_TRACKER_FEATURE_POLYLINE_2D_H_
#define PROCESSOR_TRACKER_FEATURE_POLYLINE_2D_H_

#include "laser/processor/polyline_2D_utils.h"
#include "laser/landmark/landmark_polyline_2D.h"
#include "laser/landmark/landmark_match_polyline_2D.h"
#include "laser/sensor/sensor_laser_2D.h"
#include "laser/capture/capture_laser_2D.h"
#include "laser/feature/feature_polyline_2D.h"
#include "laser/feature/feature_match_polyline_2D.h"
#include "laser/factor/factor_point_2D.h"
#include "laser/factor/factor_point_to_line_2D.h"

#include "core/processor/processor_tracker_feature.h"
#include "core/utils/params_server.hpp"

//laser_scan_utils
#include "laser_scan_utils/laser_scan.h"
#include "laser_scan_utils/line_finder_iterative.h"
#include "laser_scan_utils/polyline.h"

namespace wolf
{

WOLF_STRUCT_PTR_TYPEDEFS(ProcessorParamsTrackerFeaturePolyline2D);
WOLF_PTR_TYPEDEFS(ProcessorTrackerFeaturePolyline2D);
typedef std::list<FeatureMatchPolyline2DPtr> FeatureMatchPolyline2DPtrList;
typedef std::list<LandmarkMatchPolyline2DPtr> LandmarkMatchPolyline2DPtrList;
typedef std::map<Scalar,FeatureMatchPolyline2DPtr> FeatureMatchPolyline2DScalarMap;
typedef std::map<Scalar,LandmarkMatchPolyline2DPtr> LandmarkMatchPolyline2DScalarMap;

struct ProcessorParamsTrackerFeaturePolyline2D : public ProcessorParamsTrackerFeature
{
        ProcessorParamsTrackerFeaturePolyline2D() = default;
        ProcessorParamsTrackerFeaturePolyline2D(std::string _unique_name, const wolf::ParamsServer & _server) :
            ProcessorParamsTrackerFeature(_unique_name, _server)
        {
            // TODO write parser!
        }
        std::string print()
        {
            return "\n" + ProcessorParamsTrackerFeature::print();
            // TODO write printer!
        }

    laserscanutils::LineFinderIterativeParams line_finder_params;
    Scalar match_feature_position_sq_norm_max;
    Scalar match_feature_orientation_sq_norm_max;
    int    match_feature_overlap_n_min;
    int    match_feature_overlap_n_defined_min;
    Scalar match_feature_overlap_dist_min;

    Scalar match_landmark_position_sq_norm_max;
    Scalar match_landmark_orientation_sq_norm_max;
    int    match_landmark_overlap_n_min;
    int    match_landmark_overlap_n_defined_min;
    Scalar match_landmark_overlap_dist_min;

    bool use_probabilistic_match;

    unsigned int new_features_th;
    Scalar loop_time_th;

    std::vector<PolylineRectangularClass> polyline_classes;
    Scalar class_position_error_th;
};

class ProcessorTrackerFeaturePolyline2D : public ProcessorTrackerFeature
{
    protected:
        ProcessorParamsTrackerFeaturePolyline2DPtr params_tracker_feature_polyline_;
        laserscanutils::LineFinderIterative line_finder_;

        FeatureBasePtrList untracked_features_incoming_, untracked_features_last_;
        LandmarkMatchPolyline2DMap landmark_match_map_;
        LandmarkPolyline2DPtrList   modified_lmks_;
        std::set<LandmarkBasePtr> new_landmarks_;
        std::list<LandmarkPolyline2DPtrList> merge_candidates_list_;

        Eigen::Matrix2s R_sensor_world_last_, R_world_sensor_last_;
        Eigen::Vector2s t_sensor_world_last_, t_world_sensor_last_;
        Eigen::Matrix2s R_sensor_world_incoming_, R_world_sensor_incoming_;
        Eigen::Vector2s t_sensor_world_incoming_, t_world_sensor_incoming_;
        Eigen::Matrix2s R_robot_sensor_;
        Eigen::Vector2s t_robot_sensor_;
        Eigen::Matrix2s R_last_incoming_, R_incoming_last_;
        Eigen::Vector2s t_last_incoming_, t_incoming_last_;
        bool extrinsics_transformation_computed_;

    public:
        ProcessorTrackerFeaturePolyline2D(ProcessorParamsTrackerFeaturePolyline2DPtr _params);
        WOLF_PROCESSOR_CREATE(ProcessorTrackerFeaturePolyline2D, ProcessorParamsTrackerFeaturePolyline2D);

        virtual ~ProcessorTrackerFeaturePolyline2D();

        virtual void configure(SensorBasePtr _sensor) override {};

    protected:

        /** \brief Track provided features in \b _capture
         * \param _features_in input list of features in \b last to track
         * \param _capture the capture in which the _features_in should be searched
         * \param _features_out returned list of features found in \b _capture
         * \param _feature_correspondences returned map of correspondences: _feature_correspondences[feature_out_ptr] = feature_in_ptr
         *
         * \return the number of features tracked
         */
        virtual unsigned int trackFeatures(const FeatureBasePtrList& _features_in,
                                           const CaptureBasePtr& _capture,
                                           FeatureBasePtrList& _features_out,
                                           FeatureMatchMap& _feature_correspondences) override;

        /** \brief Correct the drift in incoming feature by re-comparing against the corresponding feature in origin.
         * \param _origin_feature input feature in origin capture tracked
         * \param _incoming_feature input/output feature in incoming capture to be corrected
         * \return false if the the process discards the correspondence with origin's feature
         */
        virtual bool correctFeatureDrift(const FeatureBasePtr _origin_feature,
                                         const FeatureBasePtr _last_feature,
                                         FeatureBasePtr _incoming_feature) override;

        /** \brief Vote for KeyFrame generation
         *
         * If a KeyFrame criterion is validated, this function returns true,
         * meaning that it wants to create a KeyFrame at the \b last Capture.
         *
         * WARNING! This function only votes! It does not create KeyFrames!
         */
        virtual bool voteForKeyFrame() override;

        /**\brief Process new Features
         *
         */
        virtual unsigned int processNew(const int& _max_features) override;

        /** \brief Detect new Features
         * \param _max_features maximum number of features detected (-1: unlimited. 0: none)
         * \param _capture The capture in which the new features should be detected.
         * \param _features_out The list of detected Features in _capture.
         * \return The number of detected Features.
         *
         * This function detects Features that do not correspond to known Features/Landmarks in the system.
         *
         * The function is called in processNew() to set the member new_features_last_,
         * the list of newly detected features of the capture last_ptr_.
         */
        virtual unsigned int detectNewFeatures(const int& _max_new_features,
                                               const CaptureBasePtr& _capture,
                                               FeatureBasePtrList& _features_out) override;

        /** \brief Emplaces a new factor
         * \param _feature_ptr pointer to the parent Feature
         * \param _feature_other_ptr pointer to the other feature constrained.
         *
         * Implement this method in derived classes.
         *
         * This function emplaces a factor of the appropriate type for the derived processor.
         */
        virtual FactorBasePtr emplaceFactor(FeatureBasePtr _feature_ptr, FeatureBasePtr _feature_other_ptr) override { return nullptr; };

        /** \brief Establish factors between features in Captures \b last and \b origin
         */
        virtual void establishFactors() override;

        /** \brief Emplaces a new point to line factor
         * \param _polyline_feature pointer to the feature
         * \param _polyline_landmark pointer to the landmark
         * \param _ftr_point_id index of the feature point
         * \param _lmk_point_id intex of the landmark point
         * \param _lmk_prev_point_id index of the landmark previous point
         */
        void emplaceFactorPointToLine(FeaturePolyline2DPtr _polyline_feature,
                                      LandmarkPolyline2DPtr _polyline_landmark,
                                      const int& _ftr_point_id,
                                      const int& _lmk_point_id,
                                      const int& _lmk_prev_point_id);

        /** \brief Emplaces a new point to point factor
         * \param _polyline_feature pointer to the feature
         * \param _polyline_landmark pointer to the landmark
         * \param _ftr_point_id index of the feature point
         * \param _lmk_point_id intex of the landmark point
         */
        void emplaceFactorPoint(FeaturePolyline2DPtr _polyline_feature,
                                LandmarkPolyline2DPtr _polyline_landmark,
                                const int& _ftr_point_id,
                                const int& _lmk_point_id);

        /** \brief emplace a landmark corresponding to a new feature
         *
         */
        virtual LandmarkBasePtr emplaceLandmark(FeatureBasePtr _feature_ptr);

        /** \brief Modify a landmark and a landmark match according to a matched feature
         * \param lmk_match landmark match
         * \param pl_ftr feature
         *
         * Grow and/or define the extremes the landmark, according to the feature.
         *
         * \return if any modification was performed
         */
        bool modifyLandmarkAndMatch(LandmarkMatchPolyline2DPtr&);

        /** \brief advance pointers
         *
         */
        virtual void advanceDerived() override;

        /** \brief reset pointers and match lists at KF creation
         *
         */
        virtual void resetDerived() override;

        /** \brief Pre-process
         *
         */
        virtual void preProcess() override;

        /** \brief Post-process
         *
         */
        virtual void postProcess() override;

        void tryMatchWithFeature(FeatureMatchPolyline2DScalarMap& ftr_matches,
                                 const FeaturePolyline2DPtr& _ftr_last,
                                 const FeaturePolyline2DPtr& _ftr_incoming,
                                 const Eigen::Matrix3s& _T_last_incoming_prior) const;

        void tryMatchWithLandmark(LandmarkMatchPolyline2DScalarMap& lmk_matches,
                                  const LandmarkPolyline2DPtr& _lmk_ptr,
                                  const FeaturePolyline2DPtr& _feat_ptr,
                                  const Eigen::Matrix3s& _T_feature_landmark_prior) const;

        bool updateLandmarkMatch(LandmarkMatchPolyline2DPtr&) const;

        bool tryUpdateMatchTransformation(LandmarkMatchPolyline2DPtr&) const;

        bool tryCompletePartialMatch(LandmarkMatchPolyline2DPtr& lmk_inc_match) const;

        void computeTransformations();

        /** \brief Concatenate feature_incoming-feature_last match with feature_last-landmark match
         * \param pl_incoming polyline feature
         * \param ftr_match_incoming_last match feature_incoming-feature_last
         * \param lmk_match_last match feature_last-landmark
         *
         * \return the concatenated match: feature_incoming-landmark
         */
        LandmarkMatchPolyline2DPtr concatenateFeatureLandmarkMatches(FeatureMatchPolyline2DPtr ftr_match_incoming_last,
                                                                     LandmarkMatchPolyline2DPtr lmk_match_last) const;

    public:

        const FeatureBasePtrList& getLastNewFeatures() const
        {
            return untracked_features_last_;
        }

        const FeatureBasePtrList& getIncomingNewFeatures() const
        {
            return untracked_features_incoming_;
        }

        const FeatureBasePtrList& getIncomingKnownFeatures() const
        {
            return known_features_incoming_;
        }
};

} /* namespace wolf */

#endif /* PROCESSOR_TRACKER_FEATURE_POLYLINE_2D_H_ */
