// sensor_laser_2D.h

#ifndef SENSOR_LASER_2D_H_
#define SENSOR_LASER_2D_H_

//wolf
#include "core/sensor/sensor_base.h"

//laser_scan_utils
#include "laser_scan_utils/laser_scan.h"

// std

namespace wolf {

WOLF_STRUCT_PTR_TYPEDEFS(IntrinsicsLaser2D);

struct IntrinsicsLaser2D : public IntrinsicsBase
{
        virtual ~IntrinsicsLaser2D() = default;

        laserscanutils::LaserScanParams scan_params;

        IntrinsicsLaser2D() = default;
        IntrinsicsLaser2D(std::string _unique_name, const wolf::ParamsServer & _server) :
            IntrinsicsBase(_unique_name, _server)
        {
            scan_params.angle_min_      = _server.getParam<double>(_unique_name + "/LaserScanParams/angle_min");
            scan_params.angle_max_      = _server.getParam<double>(_unique_name + "/LaserScanParams/angle_max");
            scan_params.angle_step_     = _server.getParam<double>(_unique_name + "/LaserScanParams/angle_step");
            scan_params.scan_time_      = _server.getParam<double>(_unique_name + "/LaserScanParams/scan_time");
            scan_params.range_min_      = _server.getParam<double>(_unique_name + "/LaserScanParams/range_min");
            scan_params.range_max_      = _server.getParam<double>(_unique_name + "/LaserScanParams/range_max");
            scan_params.range_std_dev_  = _server.getParam<double>(_unique_name + "/LaserScanParams/range_std_dev");
            scan_params.angle_std_dev_  = _server.getParam<double>(_unique_name + "/LaserScanParams/angle_std_dev");
        }
        std::string print()
        {
            return "\n" + IntrinsicsBase::print()                                                   + "\n"
            + "LaserScanParams/angle_min: "         + std::to_string(scan_params.angle_min_)        + "\n"
            + "LaserScanParams/angle_max: "         + std::to_string(scan_params.angle_max_)        + "\n"
            + "LaserScanParams/angle_step: "        + std::to_string(scan_params.angle_step_)       + "\n"
            + "LaserScanParams/scan_time: "         + std::to_string(scan_params.scan_time_)        + "\n"
            + "LaserScanParams/range_min: "         + std::to_string(scan_params.range_min_)        + "\n"
            + "LaserScanParams/range_max: "         + std::to_string(scan_params.range_max_)        + "\n"
            + "LaserScanParams/range_std_dev: "     + std::to_string(scan_params.range_std_dev_)    + "\n"
            + "LaserScanParams/angle_std_dev: "     + std::to_string(scan_params.angle_std_dev_)    + "\n";
        }

};

WOLF_PTR_TYPEDEFS(SensorLaser2D);

class SensorLaser2D : public SensorBase
{
    protected:
        laserscanutils::LaserScanParams scan_params_;

    public:
        /** \brief Constructor with extrinsics
         *
         * \param _p_ptr StateBlock pointer to the sensor position
         * \param _o_ptr StateBlock pointer to the sensor orientation
         *
         **/
        SensorLaser2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr);

        /** \brief Constructor with extrinsics and scan parameters
         *
         * \param _p_ptr StateBlock pointer to the sensor position
         * \param _o_ptr StateBlock pointer to the sensor orientation
         *
         **/
        SensorLaser2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, const double& _angle_min, const double& _angle_max, const double& _angle_step, const double& _scan_time, const double& _range_min, const double& _range_max, const double& _range_std_dev, const double& _angle_std_dev);

        /** \brief Constructor with extrinsics and scan parameters
         *
         * \param _p_ptr StateBlock pointer to the sensor position
         * \param _o_ptr StateBlock pointer to the sensor orientation
         * \param _params Scan parameters
         *
         **/
        SensorLaser2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, const laserscanutils::LaserScanParams& _params);
        SensorLaser2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, const IntrinsicsLaser2D& _params);
        SensorLaser2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, IntrinsicsLaser2DPtr _params);
        SensorLaser2D(const Eigen::VectorXs& _extrinsics, IntrinsicsLaser2DPtr _params);
        WOLF_SENSOR_CREATE(SensorLaser2D, IntrinsicsLaser2D, 3);

        virtual ~SensorLaser2D();

        void setDefaultScanParams();

        /** \brief Set scanner intrinsic parameters
         *
         * \param _params struct with scanner intrinsic parameters. See laser_scan_utils library API for reference.
         *
         **/
        void setScanParams(const laserscanutils::LaserScanParams & _params);

        /** \brief Get scanner intrinsic parameters
         *
         * Get scanner intrinsic parameters
         *
         **/
        const laserscanutils::LaserScanParams & getScanParams() const;

    public:
//        static IntrinsicsBasePtr createParams(const std::string& _filename_dot_yaml);

};

} // namespace wolf

#endif /*SENSOR_LASER_2D_H_*/
