/**
 * \file processor_odom_ICP_yaml.cpp
 *
 *  Created on: Aug 6, 2019
 *      \author: jsola
 */


// wolf
#include "laser/processor/processor_odom_icp.h"
#include "core/common/factory.h"

// wolf yaml
#include "core/yaml/yaml_conversion.h"

// yaml library
#include <yaml-cpp/yaml.h>

namespace wolf
{
namespace {
// intrinsics creator
ProcessorParamsBasePtr createProcessorParamsOdomICP(const std::string& _filename_dot_yaml)
{
    WOLF_INFO("ProcessorParamsOdomICP: Parsing file: ", _filename_dot_yaml);

    YAML::Node config = YAML::LoadFile(_filename_dot_yaml);

    if (config["type"].as<std::string>() == "ODOM ICP")
    {
        ProcessorParamsOdomICPPtr params = std::make_shared<ProcessorParamsOdomICP>();

        params->voting_active               = config["voting_active"]               .as<bool>();
        params->voting_aux_active           = config["voting_aux_active"]           .as<bool>();
        params->time_tolerance              = config["time_tolerance"]              .as<Scalar>();

        params->min_features_for_keyframe   = config["min_features_for_keyframe"]   .as<Scalar>();
        params->max_new_features            = config["max_new_features"]            .as<SizeEigen  >();

        params->use_point_to_line_distance  = config["use_point_to_line_distance"]  .as<int>();
        params->max_correspondence_dist     = config["max_correspondence_dist"]     .as<int>();
        params->max_iterations              = config["max_iterations"]              .as<int>();
        params->use_corr_tricks             = config["use_corr_tricks"]             .as<int>();
        params->outliers_maxPerc            = config["outliers_maxPerc"]            .as<Scalar>();
        params->outliers_adaptive_order     = config["outliers_adaptive_order"]     .as<Scalar>();
        params->outliers_adaptive_mult      = config["outliers_adaptive_mult"]      .as<Scalar>();
        params->vfk_min_dist                = config["vfk_min_dist"]                .as<Scalar>();
        params->vfk_min_angle               = config["vfk_min_angle"]               .as<Scalar>();
        params->vfk_min_time                = config["vfk_min_time"]                .as<Scalar>();
        params->vfk_min_error               = config["vfk_min_error"]               .as<Scalar>();
        params->vfk_max_points              = config["vfk_max_points"]              .as<int>();

        return params;
    }

    std::cout << "Bad configuration file. No processor type found." << std::endl;
    return nullptr;
}

// register into factory
const bool WOLF_UNUSED registered_odom_ICP_params = ProcessorParamsFactory::get().registerCreator("ODOM ICP", createProcessorParamsOdomICP);

} // namespace [void]
} // namespace wolf
