/**
 * \file sensor_laser_2D_yaml.cpp
 *
 *  Created on: May 13, 2016
 *      \author: jsola
 */

// wolf yaml
#include "core/yaml/yaml_conversion.h"

// wolf
//#include "core/intrinsics_factory.h"
#include "core/common/factory.h"
#include "laser/sensor/sensor_laser_2D.h"

// yaml library
#include <yaml-cpp/yaml.h>

namespace wolf
{
namespace {
// intrinsics creator
IntrinsicsBasePtr createIntrinsicsLaser2D(const std::string& _filename_dot_yaml)
{
    // If required: Parse YAML

    WOLF_INFO("IntrinsicsLaser2D: Parsing file: ", _filename_dot_yaml);

    YAML::Node config = YAML::LoadFile(_filename_dot_yaml);

    if (config["type"].as<std::string>() == "LASER 2D")
    {
        YAML::Node ls_params = config["LaserScanParams"];

        IntrinsicsLaser2DPtr params = std::make_shared<IntrinsicsLaser2D>();

        params->scan_params.angle_min_      = ls_params["angle_min"]      .as<double>();
        params->scan_params.angle_max_      = ls_params["angle_max"]      .as<double>();
        params->scan_params.angle_step_     = ls_params["angle_step"]     .as<double>();
        params->scan_params.scan_time_      = ls_params["scan_time"]      .as<double>();
        params->scan_params.range_min_      = ls_params["range_min"]      .as<double>();
        params->scan_params.range_max_      = ls_params["range_max"]      .as<double>();
        params->scan_params.range_std_dev_  = ls_params["range_std_dev"]  .as<double>();
        params->scan_params.angle_std_dev_  = ls_params["angle_std_dev"]  .as<double>();

        return params;
    }

    std::cout << "Bad configuration file " << _filename_dot_yaml << ". No processor type found." << std::endl;
    return nullptr;

}

// register into factory
const bool WOLF_UNUSED registered_laser_params = IntrinsicsFactory::get().registerCreator("LASER 2D", createIntrinsicsLaser2D);

} // namespace [void]
} // namespace wolf
